module std.atomic;


/**
 * Memory synchronization flag.  If the supplied option is not available on the
 * current platform then a stronger method will be used instead.
 */
enum msync
{
    hlb,    /// hoist-load barrier
    ddhlb,  /// hoist-load barrier with data-dependency "hint"
    hsb,    /// hoist-store barrier
    slb,    /// sink-load barrier
    ssb,    /// sink-store barrier
    acq,    /// hoist-load + hoist-store barrier
    rel,    /// sink-load + sink-store barrier
    none    /// naked
}


private
{
    template isValidAtomicType( T )
    {
        const bit isValidAtomicType = is(T:bit)   ||
                                      is(T:byte)  || is(T:ubyte)  ||
                                      is(T:short) || is(T:ushort) ||
                                      is(T:int)   || is(T:uint)   ||
                                      is(T:long)  || is(T:ulong);
    }

    template isValidAtomicType( T : T* )
    {
        const bit isValidAtomicType = true;
    }

    template suppliedSizeofMatchesTrueSizeof( T, size_t sizeofT )
    {
        const bit suppliedSizeofMatchesTrueSizeof = sizeofT == T.sizeof;
    }

    template isHoistOp( msync ms )
    {
        const bit isHoistOp = ms == msync.hlb   ||
                              ms == msync.ddhlb ||
                              ms == msync.hsb   ||
                              ms == msync.acq;
    }

    template isSinkOp( msync ms )
    {
        const bit isSinkOp = ms == msync.slb ||
                             ms == msync.ssb ||
                             ms == msync.rel;
    }

    version( X86 )
    {
        template atomicValueIsProperlyAligned( T )
        {
            bit atomicValueIsProperlyAligned( size_t addr )
            {
                return addr % T.sizeof == 0;
            }
        }

        template doAtomicLoad( T, bit membar, size_t sizeofT : 1 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, val;
                        lock;
                        mov AL, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }

        template doAtomicLoad( T, bit membar, size_t sizeofT : 2 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, val;
                        lock;
                        mov AX, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }

        template doAtomicLoad( T, bit membar, size_t sizeofT : 4 = T.sizeof )
        {
            T doAtomicLoad( inout T val )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, val;
                        lock;
                        mov EAX, [EAX];
                    }
                }
                else
                {
                    volatile
                    {
                        return val;
                    }
                }
            }
        }

        version( X86_64 )
        {
            template doAtomicLoad( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicLoad( inout T val )
                in
                {
                    bit isProperlyAligned = cast(size_t) &val % val.sizeof == 0;
                    assert( isProperlyAligned );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            lock;
                            mov RAX, [RAX];
                        }
                    }
                    else
                    {
                        volatile
                        {
                            return val;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicLoad( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                T doAtomicLoad( inout T val )
                {
                    const bit operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }

        template doAtomicStore( T, bit membar, size_t sizeofT : 1 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, pval;
                        mov BL, newval;
                        lock;
                        xchg [EAX], BL;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov EAX, pval;
                        mov BL, newval;
                        mov [EAX], BL;
                    }
                }
            }
        }

        template doAtomicStore( T, bit membar, size_t sizeofT : 2 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, pval;
                        mov BX, newval;
                        lock;
                        xchg [EAX], BX;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov EAX, pval;
                        mov BX, newval;
                        mov [EAX], BX;
                    }
                }
            }
        }

        template doAtomicStore( T, bit membar, size_t sizeofT : 4 = T.sizeof )
        {
            void doAtomicStore( inout T val, T newval )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EAX, val;
                        mov EBX, newval;
                        lock;
                        xchg [EAX], EBX;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov EAX, val;
                        mov EBX, newval;
                        mov [EAX], EBX;
                    }
                }
            }
        }

        version( X86_64 )
        {
            template doAtomicStore( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                void doAtomicStore( inout T val, T newval )
                in
                {
                    bit isProperlyAligned = cast(size_t) &val % val.sizeof == 0;
                    assert( isProperlyAligned );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            lock;
                            xchg [RAX], RBX;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            mov [RAX], RBX;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicStore( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                void doAtomicStore( inout T val, T newval )
                {
                    const bit operationOnlyAvailableFor64bit = false;
                    static assert( operationOnlyAvailableFor64bit );
                }
            }
        }

        template doAtomicStoreIf( T, bit membar, size_t sizeofT : 1 = T.sizeof )
        {
            bit doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov BL, newval;
                        mov AL, equalTo;
                        mov ECX, pval;
                        lock;
                        cmpxchg [ECX], BL;
                        setz AL;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov BL, newval;
                        mov AL, equalTo;
                        mov ECX, pval;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], BL;
                        setz AL;
                    }
                }
            }
        }

        template doAtomicStoreIf( T, bit membar, size_t sizeofT : 2 = T.sizeof )
        {
            bit doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                T* pval = &val;
                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov BX, newval;
                        mov AX, equalTo;
                        mov ECX, pval;
                        lock;
                        cmpxchg [ECX], BX;
                        setz AL;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov BX, newval;
                        mov AX, equalTo;
                        mov ECX, pval;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], BX;
                        setz AL;
                    }
                }
            }
        }

        template doAtomicStoreIf( T, bit membar, size_t sizeofT : 4 = T.sizeof )
        {
            bit doAtomicStoreIf( inout T val, T newval, T equalTo )
            in
            {
                assert( atomicValueIsProperlyAligned!(T)( cast(size_t) &val ) );
            }
            body
            {
                static assert( isValidAtomicType!(T) );
                static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                static if( membar )
                {
                    /*volatile*/ asm
                    {
                        mov EBX, newval;
                        mov EAX, equalTo;
                        mov ECX, val;
                        lock;
                        cmpxchg [ECX], EBX;
                        setz AL;
                    }
                }
                else
                {
                    /*volatile*/ asm
                    {
                        mov EBX, newval;
                        mov EAX, equalTo;
                        mov ECX, val;
                        lock; // lock needed to make this op atomic
                        cmpxchg [ECX], EBX;
                        setz AL;
                    }
                }
            }
        }

        version( X86_64 )
        {
            template doAtomicStoreIf( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                bit doAtomicStoreIf( inout T val, T newval, T equalTo )
                in
                {
                    bit isProperlyAligned = cast(size_t) &val % val.sizeof == 0;
                    assert( isProperlyAligned );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock;
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock; // lock needed to make this op atomic
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicStoreIf( T, bit membar, size_t sizeofT : 8 = T.sizeof )
            {
                bit doAtomicStoreIf( inout T val, T newval, T equalTo )
                in
                {
                    bit isProperlyAligned = cast(size_t) &val % val.sizeof == 0;
                    assert( isProperlyAligned );
                }
                body
                {
                    static assert( isValidAtomicType!(T) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            lea	EDI, newval;
                            mov	EBX, [EDI];
                            mov	ECX, 4[EDI];
                            lea	EDI, equalTo;
                            mov	EAX, [EDI];
                            mov	EDX, 4[EDI];
                            mov	EDI, val;
                            lock;
                            cmpxch8b [EDI];
                            setz AL;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            lea	EDI, newval;
                            mov	EBX, [EDI];
                            mov	ECX, 4[EDI];
                            lea	EDI, equalTo;
                            mov	EAX, [EDI];
                            mov	EDX, 4[EDI];
                            mov	EDI, val;
                            lock; // lock needed to make this op atomic
                            cmpxch8b [EDI];
                            setz AL;
                        }
                    }
                }
            }
        }

        //
        // NOTE: The following specializations are temporary until DMD is fixed
        //       to correctly handle multiple same-sized types (as with int/uint).
        //
        version( X86_64 )
        {
            template doAtomicLoad( T : T*, bit membar, size_t sizeofT : 8 = (T*).sizeof )
            {
                T* doAtomicLoad( inout T* val )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            lock;
                            mov RAX, [RAX];
                        }
                    }
                    else
                    {
                        volatile
                        {
                            return val;
                        }
                    }
                }
            }

            template doAtomicStore( T : T*, bit membar, size_t sizeofT : 8 = (T*).sizeof )
            {
                void doAtomicStore( inout T* val, T* newval )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            lock;
                            xchg [RAX], RBX;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov RAX, val;
                            mov RBX, newval;
                            mov [RAX], RBX;
                        }
                    }
                }
            }

            template doAtomicStoreIf( T : T*, bit membar, size_t sizeofT : 4 = (T*).sizeof )
            {
                bit doAtomicStoreIf( inout T* val, T* newval, T* equalTo )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock;
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov RBX, newval;
                            mov RAX, equalTo;
                            mov RCX, val;
                            lock; // lock needed to make this op atomic
                            cmpxchg [RCX], RBX;
                            setz AL;
                        }
                    }
                }
            }
        }
        else
        {
            template doAtomicLoad( T : T*, bit membar, size_t sizeofT : 4 = (T*).sizeof )
            {
                T* doAtomicLoad( inout T* val )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov EAX, val;
                            lock;
                            mov EAX, [EAX];
                        }
                    }
                    else
                    {
                        volatile
                        {
                            return val;
                        }
                    }
                }
            }

            template doAtomicStore( T : T*, bit membar, size_t sizeofT : 4 = (T*).sizeof )
            {
                void doAtomicStore( inout T* val, T* newval )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov EAX, val;
                            mov EBX, newval;
                            lock;
                            xchg [EAX], EBX;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov EAX, val;
                            mov EBX, newval;
                            mov [EAX], EBX;
                        }
                    }
                }
            }

            template doAtomicStoreIf( T : T*, bit membar, size_t sizeofT : 4 = (T*).sizeof )
            {
                bit doAtomicStoreIf( inout T* val, T* newval, T* equalTo )
                in
                {
                    assert( atomicValueIsProperlyAligned!(T*)( cast(size_t) &val ) );
                }
                body
                {
                    static assert( isValidAtomicType!(T*) );
                    static assert( suppliedSizeofMatchesTrueSizeof!( T*, sizeofT ) );

                    static if( membar )
                    {
                        /*volatile*/ asm
                        {
                            mov EBX, newval;
                            mov EAX, equalTo;
                            mov ECX, val;
                            lock;
                            cmpxchg [ECX], EBX;
                            setz AL;
                        }
                    }
                    else
                    {
                        /*volatile*/ asm
                        {
                            mov EBX, newval;
                            mov EAX, equalTo;
                            mov ECX, val;
                            lock; // lock needed to make this op atomic
                            cmpxchg [ECX], EBX;
                            setz AL;
                        }
                    }
                }
            }
        }
    }
}

version( X86 )
{
    //
    // NOTE: x86 loads implicitly have acquire semantics so a membar is only necessary on release
    //

    /**
     * Refreshes the contents of 'val' from main memory.  This operation is both lock-free and atomic.
     *
     * Returns:
     *  The loaded value.
     *
     * Params:
     *  val = The value to load.  This value must be properly aligned.
     */
    template atomicLoad( T, msync ms : msync.none )  { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; }
    template atomicLoad( T, msync ms : msync.hlb )   { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto
    template atomicLoad( T, msync ms : msync.ddhlb ) { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto
    template atomicLoad( T, msync ms : msync.acq )   { alias doAtomicLoad!(T,isSinkOp!(ms)) atomicLoad; } /// ditto

    //
    // NOTE: x86 stores implicitly have release semantics so a membar is only necessary on acquires
    //

    /**
     * Stores 'newval' to the memory referenced by 'val'.  This operation is both lock-free and atomic.
     *
     * Params:
     *  val     = The destination variable.
     *  newval  = The value to store.
     */
    template atomicStore( T, msync ms : msync.none ) { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; }
    template atomicStore( T, msync ms : msync.ssb )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto
    template atomicStore( T, msync ms : msync.acq )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto
    template atomicStore( T, msync ms : msync.rel )  { alias doAtomicStore!(T,isHoistOp!(ms)) atomicStore; } /// ditto

    /**
     * Stores 'newval' to the memory referenced by 'val' if val is equal to 'equalTo'.  This operation
     * is both lock-free and atomic.
     *
     * Returns:
     *  true if the store occurred, false if not.
     *
     * Params:
     *  val     = The destination variable.
     *  newval  = The value to store.
     *  equalTo = The comparison value.
     */
    template atomicStoreIf( T, msync ms : msync.none ) { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; }
    template atomicStoreIf( T, msync ms : msync.ssb )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto
    template atomicStoreIf( T, msync ms : msync.acq )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto
    template atomicStoreIf( T, msync ms : msync.rel )  { alias doAtomicStoreIf!(T,ms!=msync.none) atomicStoreIf; } /// ditto
}

/**
 * This class represents a value which will be subject to competing access.  All accesses to
 * this value will be synchronized with main memory, and various memory barriers may be
 * employed for instruction ordering.  Any primitive type of size equal to or smaller than
 * the memory bus size is allowed, so 32-bit machines may use values with size <= int.sizeof
 * and 64-bit machines may use values with size <= long.sizeof.  The one exception to this
 * rule is that architectures that support DCAS will allow double-wide storeIf operations.
 * The 32-bit x86 architecture, for example, supports 64-bit storeIf operations.
 */
struct Atomic( T )
{
    static assert( isValidAtomicType!(T) );
/+
    template load( msync ms : msync.none )  { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.hlb )   { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.ddhlb ) { alias atomicLoad!(T,ms) load; }
    template load( msync ms : msync.acq )   { alias atomicLoad!(T,ms) load; }

    template store( msync ms : msync.none ) { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.ssb )  { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.acq )  { alias atomicStore!(T,ms) store; }
    template store( msync ms : msync.rel )  { alias atomicStore!(T,ms) store; }

    template storeIf( msync ms : msync.none ) { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.ssb )  { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.acq )  { alias atomicStoreIf!(T,ms) store; }
    template storeIf( msync ms : msync.rel )  { alias atomicStoreIf!(T,ms) store; }
+/
    /**
     * Refreshes the contents of this value from main memory.  This operation is both lock-free and atomic.
     *
     * Returns:
     *  The loaded value.
     */
    template load( msync ms : msync.none )  { T load() { return atomicLoad!(T,ms)( m_val ); } }
    template load( msync ms : msync.hlb )   { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto
    template load( msync ms : msync.ddhlb ) { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto
    template load( msync ms : msync.acq )   { T load() { return atomicLoad!(T,ms)( m_val ); } } /// ditto

    /**
     * Stores 'newval' to the memory referenced by this value.  This operation is both lock-free and atomic.
     *
     * Params:
     *  newval  = The value to store.
     */
    template store( msync ms : msync.none ) { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } }
    template store( msync ms : msync.ssb )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto
    template store( msync ms : msync.acq )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto
    template store( msync ms : msync.rel )  { void store( T newval ) { atomicStore!(T,ms)( m_val, newval ); } } /// ditto

    /**
     * Stores 'newval' to the memory referenced by this value if val is equal to 'equalTo'.  This operation
     * is both lock-free and atomic.
     *
     * Returns:
     *  true if the store occurred, false if not.
     *
     * Params:
     *  newval  = The value to store.
     *  equalTo = The comparison value.
     */
    template storeIf( msync ms : msync.none ) { bit storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } }
    template storeIf( msync ms : msync.ssb )  { bit storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto
    template storeIf( msync ms : msync.acq )  { bit storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto
    template storeIf( msync ms : msync.rel )  { bit storeIf( T newval, T equalTo ) { return atomicStoreIf!(T,ms)( m_val, newval, equalTo ); } } /// ditto

private:
    T   m_val;
}

private
{
    template testLoad( T, msync ms )
    {
        void testLoad()
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.load!(ms)() == base );
            base        = 1;
            atom.m_val  = 1;
            assert( atom.load!(ms)() == base );
        }
    }

    template testStore( T, msync ms )
    {
        void testStore()
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.m_val == base );
            base = 1;
            atom.store!(ms)( base );
            assert( atom.m_val == base );
        }
    }

    template testStoreIf( T, msync ms )
    {
        void testStoreIf()
        {
            T          base;
            Atomic!(T) atom;

            assert( atom.m_val == base );
            base = 1;
            atom.storeIf!(ms)( base, 1 );
            assert( atom.m_val != base );
            atom.storeIf!(ms)( base, T.init );
            assert( atom.m_val == base );
        }
    }

    template testType( T )
    {
        void testType()
        {
            testLoad!(T, msync.none)();
            testLoad!(T, msync.acq)();

            testStore!(T, msync.none)();
            testStore!(T, msync.acq)();
            testStore!(T, msync.rel)();

            testStoreIf!(T, msync.none)();
            testStoreIf!(T, msync.acq)();
        }
    }
}

unittest
{
    //
    // byte
    //

    testLoad!(byte, msync.none)();
    testLoad!(byte, msync.acq)();

    testStore!(byte, msync.none)();
    testStore!(byte, msync.acq)();
    testStore!(byte, msync.rel)();

    testStoreIf!(byte, msync.none)();
    testStoreIf!(byte, msync.acq)();
/+
    //
    // ubyte
    //

    testLoad!(ubyte, msync.none)();
    testLoad!(ubyte, msync.acq)();

    testStore!(ubyte, msync.none)();
    testStore!(ubyte, msync.acq)();
    testStore!(ubyte, msync.rel)();

    testStoreIf!(ubyte, msync.none)();
    testStoreIf!(ubyte, msync.acq)();
+/
    //
    // short
    //

    testLoad!(short, msync.none)();
    testLoad!(short, msync.acq)();

    testStore!(short, msync.none)();
    testStore!(short, msync.acq)();
    testStore!(short, msync.rel)();

    testStoreIf!(short, msync.none)();
    testStoreIf!(short, msync.acq)();
/+
    //
    // ushort
    //

    testLoad!(ushort, msync.none)();
    testLoad!(ushort, msync.acq)();

    testStore!(ushort, msync.none)();
    testStore!(ushort, msync.acq)();
    testStore!(ushort, msync.rel)();

    testStoreIf!(ushort, msync.none)();
    testStoreIf!(ushort, msync.acq)();
+/
    //
    // int
    //

    testLoad!(int, msync.none)();
    testLoad!(int, msync.acq)();

    testStore!(int, msync.none)();
    testStore!(int, msync.acq)();
    testStore!(int, msync.rel)();

    testStoreIf!(int, msync.none)();
    testStoreIf!(int, msync.acq)();
/+
    //
    // uint
    //

    testLoad!(uint, msync.none)();
    testLoad!(uint, msync.acq)();

    testStore!(uint, msync.none)();
    testStore!(uint, msync.acq)();
    testStore!(uint, msync.rel)();

    testStoreIf!(uint, msync.none)();
    testStoreIf!(uint, msync.acq)();
+/
    //
    // long
    //

    version( X86_64 )
    {
        testLoad!(long, msync.none)();
        testLoad!(long, msync.acq)();

        testStore!(long, msync.none)();
        testStore!(long, msync.acq)();
        testStore!(long, msync.rel)();
    }

    testStoreIf!(long, msync.none)();
    testStoreIf!(long, msync.acq)();
/+
    //
    // ulong
    //

    version( X86_64 )
    {
        testLoad!(ulong, msync.none)();
        testLoad!(ulong, msync.acq)();

        testStore!(ulong, msync.none)();
        testStore!(ulong, msync.acq)();
        testStore!(ulong, msync.rel)();
    }

    testStoreIf!(ulong, msync.none)();
    testStoreIf!(ulong, msync.acq)();
+/
}