module std.exception;

private
{
    import std.c.stdio;
    import std.c.stdlib;
}

private
{
    alias void function( char[] file, uint line ) assertHandlerType;
    alias void function() terminateHandlerType;

    assertHandlerType assertHandler = null;
}

/**
 * Thrown on an array bounds error.
 */
class ArrayBoundsException : Exception
{
    this( char[] file, size_t line )
    {
        super( "array index out of bounds", file, line );
    }
}

/**
 * Thrown on an assert error.
 */
class AssertException : Exception
{
    this( char[] file, size_t line )
    {
        super( "assertion failure", file, line );
    }
}

/**
 * Thrown on an out of memory error.
 */
class OutOfMemoryException : Exception
{
    this( char[] file, size_t line )
    {
        super( "memory allocation failed", file, line );
    }
}

/**
 * Thrown on a switch error.
 */
class SwitchException : Exception
{
    this( char[] file, size_t line )
    {
        super( "no appropriate switch clause found", file, line );
    }
}

/**
 * Thrown on an invalid UTF error.
 */
class InvalidUtfException : Exception
{
    size_t idx;

    this( char[] msg, size_t idx )
    {
        super( msg );
        this.idx = idx;
    }
}

/**
 * Overrides the default assert hander with a user-supplied version.
 *
 * Params:
 *  h = The new assert handler.
 */
void setAssertHandler( assertHandlerType h )
{
    assertHandler = h;
}

/**
 * A callback for assert errors in D.  The user-supplied assert handler will
 * be called if one has been supplied, otherwise an AssertException will be
 * thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onAssert( char[] file, uint line )
{
    if( assertHandler )
        assertHandler( file, line );
    else
        throw new AssertException( file, line );
}

/**
 * A callback for out of memory errors in D.  An OutOfMemoryException will be
 * thrown.
 */
extern (C) void onOutOfMemory()
{
    throw cast(OutOfMemoryException) cast(void*) OutOfMemoryException.classinfo.init;
}

/**
 * A callback for array bounds errors in D.  An ArrayBoundsException will be
 * thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onArrayBoundsError( char[] file, size_t line )
{
    throw new ArrayBoundsException( file, line );
}

/**
 * A callback for switch errors in D.  A SwitchException will be thrown.
 *
 * Params:
 *  file = The name of the file that signaled this error.
 *  line = The line number on which this error occurred.
 */
extern (C) void onSwitchError( char[] file, size_t line )
{
    throw new SwitchException( file, line );
}

/**
 * A callback for internal UTF errors in D.  An InvalidUtfException will be
 * thrown.
 *
 * Params:
 *  msg = Information about the error.
 *  idx = String index where this error was detected.
 */
extern (C) void onInvalidUtfError( char[] msg, size_t idx )
{
    throw new InvalidUtfException( msg, idx );
}