module std.atomic;


private
{
    ////////////////////////////////////////////////////////////////////////////////
    // Utility
    ////////////////////////////////////////////////////////////////////////////////


    template LargerOf( T1, T2 )
    {
        static if( T1.sizeof >= T2.sizeof )
            alias T1 LargerOf;
        else
            alias T2 LargerOf;
    }


    ////////////////////////////////////////////////////////////////////////////////
    // Unary Operators
    ////////////////////////////////////////////////////////////////////////////////


    struct Neg ( T ) { static T opCall( T t ) { return -t; } }
    struct Pos ( T ) { static T opCall( T t ) { return +t; } }
    struct Com ( T ) { static T opCall( T t ) { return ~t; } }


    ////////////////////////////////////////////////////////////////////////////////
    // Binary Operators
    ////////////////////////////////////////////////////////////////////////////////


    struct Add ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t +   u; } }
    struct Sub ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t +   u; } }
    struct Mul ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t *   u; } }
    struct Div ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t /   u; } }
    struct Mod ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t %   u; } }
    struct And ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t &   u; } }
    struct Or  ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t |   u; } }
    struct Xor ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t ^   u; } }
    struct Shl ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t <<  u; } }
    struct Shr ( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t >>  u; } }
    struct UShr( T1, T2 = T1 ) { static LargerOf!(T1,T2) opCall( T1 t, T2 u ) { return t >>> u; } }


    ////////////////////////////////////////////////////////////////////////////////
    // ValueImpl
    ////////////////////////////////////////////////////////////////////////////////


    template ValueImpl( T )
    {
        alias T ValueType;

        T val()
        {
            return Platform!(T).load( m_val );
        }

        T val( T val )
        {
            Platform!(T).store( m_val, val );
            return val;
        }

        bit storeIf( T val, T equalTo )
        {
           return Platform!(T).storeIf( m_val, val, equalTo );
        }

        T opCast() { return  Platform!(T).load( m_val ); }
        T opNeg () { return -Platform!(T).load( m_val ); }
        T opPos () { return +Platform!(T).load( m_val ); }
        T opCom () { return ~Platform!(T).load( m_val ); }

        T opPostInc()
        {
            T ret = Platform!(T).load( m_val );
            T val = ret;
            while( !Platform!(T).storeIf( m_val, val + 1, val ) )
                val = Platform!(T).load( m_val );
            return ret;
        }

        T opPostDec()
        {
            T ret = Platform!(T).load( m_val );
            T val = ret;
            while( !Platform!(T).storeIf( m_val, val - 1, val ) )
                val = Platform!(T).load( m_val );
            return ret;
        }

        T opAdd      ( T val ) { return Platform!(T).load( m_val ) +   val; }
        T opSub      ( T val ) { return Platform!(T).load( m_val ) -   val; }
        T opMul      ( T val ) { return Platform!(T).load( m_val ) *   val; }
        T opDiv      ( T val ) { return Platform!(T).load( m_val ) /   val; }
        T opMod      ( T val ) { return Platform!(T).load( m_val ) %   val; }
        T opAnd      ( T val ) { return Platform!(T).load( m_val ) &   val; }
        T opOr       ( T val ) { return Platform!(T).load( m_val ) |   val; }
        T opXor      ( T val ) { return Platform!(T).load( m_val ) ^   val; }
        T opShl      ( T val ) { return Platform!(T).load( m_val ) <<  val; }
        T opShr      ( T val ) { return Platform!(T).load( m_val ) >>  val; }
        T opUShr     ( T val ) { return Platform!(T).load( m_val ) >>> val; }

        T opAdd_r    ( T val ) { return val +   Platform!(T).load( m_val ); }
        T opSub_r    ( T val ) { return val -   Platform!(T).load( m_val ); }
        T opMul_r    ( T val ) { return val *   Platform!(T).load( m_val ); }
        T opDiv_r    ( T val ) { return val /   Platform!(T).load( m_val ); }
        T opMod_r    ( T val ) { return val %   Platform!(T).load( m_val ); }
        T opAnd_r    ( T val ) { return val &   Platform!(T).load( m_val ); }
        T opOr_r     ( T val ) { return val |   Platform!(T).load( m_val ); }
        T opXor_r    ( T val ) { return val ^   Platform!(T).load( m_val ); }
        T opShl_r    ( T val ) { return val <<  Platform!(T).load( m_val ); }
        T opShr_r    ( T val ) { return val >>  Platform!(T).load( m_val ); }
        T opUShr_r   ( T val ) { return val >>> Platform!(T).load( m_val ); }

        int opEquals ( T val ) { return Platform!(T).load( m_val ) ==  val; }
        int opCmp    ( T val )
        {
            T tmp = Platform!(T).load( m_val );
            if( tmp < val ) return -1;
            if( tmp > val ) return  1;
            return 0;
        }

        T opAddAssign  ( T val ) { return doOp!(Add!(T))  ( val ); }
        T opSubAssign  ( T val ) { return doOp!(Sub!(T))  ( val ); }
        T opMulAssign  ( T val ) { return doOp!(Mul!(T))  ( val ); }
        T opDivAssign  ( T val ) { return doOp!(Div!(T))  ( val ); }
        T opModAssign  ( T val ) { return doOp!(Mod!(T))  ( val ); }
        T opAndAssign  ( T val ) { return doOp!(And!(T))  ( val ); }
        T opOrAssign   ( T val ) { return doOp!(Or!(T))   ( val ); }
        T opXorAssign  ( T val ) { return doOp!(Xor!(T))  ( val ); }
        T opShlAssign  ( T val ) { return doOp!(Shl!(T))  ( val ); }
        T opShrAssign  ( T val ) { return doOp!(Shr!(T))  ( val ); }
        T opUShrAssign ( T val ) { return doOp!(UShr!(T)) ( val ); }

    private:
        template doOp( Op )
        {
            T doOp( T val )
            {
                T tmp = Platform!(T).load( m_val );
                while( !Platform!(T).storeIf( m_val, Op( tmp, val ), tmp ) )
                    tmp = Platform!(T).load( m_val );
                return Op( tmp, val );
            }
        }

        // NOTE: to guarantee optimal performance, m_val should be aligned
        //       on a 32 or 64-bit boundary (depending on data size)
        T   m_val;
    }
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (int)
////////////////////////////////////////////////////////////////////////////////


struct Atomic( T : int )
{
    mixin Platform!(T);
    mixin ValueImpl!(T);
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (uint)
////////////////////////////////////////////////////////////////////////////////


struct Atomic( T : uint )
{
    mixin Platform!(T);
    mixin ValueImpl!(T);
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (long)
////////////////////////////////////////////////////////////////////////////////


struct Atomic( T : long )
{
    mixin Platform!(T);
    mixin ValueImpl!(T);
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (ulong)
////////////////////////////////////////////////////////////////////////////////


struct Atomic( T : ulong )
{
    mixin Platform!(T);
    mixin ValueImpl!(T);
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (bit)
////////////////////////////////////////////////////////////////////////////////


struct Atomic( T : bit )
{
    alias T ValueType;

    T val()
    {
        return m_val != 0;
    }

    T val( T val )
    {
        PlatformT.store( m_val, val ? 1 : 0 );
        return val != 0;
    }

    T opCast()
    {
        return PlatformT.load( m_val ) != 0;
    }

    T opEquals ( T val )
    {
        StorageT tmp = PlatformT.load( m_val );
        return val ? tmp != 0 : tmp == 0;
    }

private:
    invariant
    {
        StorageT tmp = PlatformT.load( m_val );
        assert( m_val == 0 || m_val == 1 );
    }

    alias uint                StorageT;
    alias Platform!(StorageT) PlatformT;

    uint    m_val;
}


////////////////////////////////////////////////////////////////////////////////
// Atomic (T*)
////////////////////////////////////////////////////////////////////////////////


