module std.thread;

/**
 * All exceptions thrown from this module derive from this class.
 */
class ThreadError : Exception
{
    this( char[] msg )
    {
        super( msg );
    }
}

private
{
    template min(T)
    {
        T min( T lhs, T rhs ) { return rhs < lhs ? rhs : lhs; }
    }

    template max(T)
    {
        T max( T lhs, T rhs ) { return lhs > rhs ? lhs : rhs; }
    }

    version( X86 )
    {
        enum
        {
            edi,
            esi,
            ebp,
            esp,
            ebx,
            edx,
            ecx,
            eax
        }
    }
}

version( Win32 )
{
    private
    {
        import sys.windows.c.process;
        import sys.windows.c.windows;
        import std.c.string; // for memset

        //
        // entry point for Win32 threads
        //
        extern (Windows) uint threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

            obj.m_bstack = getStackBottom();
            obj.m_tstack = obj.m_bstack;
            TlsSetValue( Thread.sm_this, obj );
	        Thread.add( obj );
	        try
	        {
	            obj.run();
	        }
	        catch
	        {
	            // error should really print to stderr
	        }
	        Thread.remove( obj );
	        return 0;
        }

        //
        // copy of the same-named function in phobos.std.thread,
        // it uses the Windows naming convention to be consistent
        // with GetCurrentThreadId
        //
        HANDLE GetCurrentThreadHandle()
        {
            const uint DUPLICATE_SAME_ACCESS = 0x00000002;

    	    HANDLE curr = GetCurrentThread(),
    	           proc = GetCurrentProcess(),
    	           hndl;

    	    DuplicateHandle( proc, curr, proc, &hndl, 0, TRUE, DUPLICATE_SAME_ACCESS );
    	    return hndl;
        }

        void* getStackBottom()
        {
            asm
            {
                naked;
                mov	EAX, FS:4;
                ret;
            }
        }

        void* getStackTop()
        {
            asm
            {
        	    naked;
                mov EAX, ESP;
                ret;
            }
        }
    }

    /**
     * This class encapsulates all threading functionality for the D
     * programming language.  As thread manipulation is a required facility
     * for garbage collection, all user threads should derive from this
     * class, and instances of this class should never be explicitly deleted.
     */
    class Thread
    {
        /**
         * Initializes a thread object which has no associated executable
         * function.
         */
        this()
        {
            m_call = Call.NO;
        }

        /**
         * Initialized a thread object which is associated with a static
         * D function.
         *
         * Params:
         *  fn = The thread function.
         */
        this( void function() fn )
        {
            m_fn   = fn;
            m_call = Call.FN;
        }

        /**
         * Initializes a thread object which is associated with a dynamic
         * D function.
         *
         * Params:
         *  dg = The thread function.
         */
        this( void delegate() dg )
        {
            m_dg   = dg;
            m_call = Call.DG;
        }

        /**
         * Performs routine cleanup for this object.
         *
         * In:
         *  assert( !isRunning() );
         */
        ~this()
        in
        {
            assert( !isRunning() );
        }
        body
        {
            if( m_addr )
                remove( this );
            CloseHandle( m_hndl );
        }

        /**
         * Gets the user-readable label for this thread.
         *
         * Returns:
         *  The name of this thread.
         */
        final char[] name()
        {
            synchronized( this )
            {
                return m_name;
            }
        }

        /**
         * Sets the user-readable label for this thread.
         *
         * Params:
         *  n = The new name for this thread.
         */
        final void name( char[] n )
        {
            synchronized( this )
            {
                m_name = n.dup;
            }
        }

        /**
         * Starts the thread with run() as the target method.  The default
         * behavior of this is to run the function or delegate passed upon
         * construction.
         *
         * In:
         *  This routine may only be called once per thread instance.
         *
         * Throws:
         *  ThreadError if the thread fails to start.
         */
        final void start()
        in
        {
            assert( !m_hndl && !m_addr );
        }
        body
        {
            multiThreadedFlag = true;
            m_hndl = cast(HANDLE) _beginthreadex( null, 0, &threadFunc, this, 0, &m_addr );
            if( cast(size_t) m_hndl == 0 )
                throw new ThreadError( "Error creating thread" );
        }

        /**
         * Tests whether this thread is running.  This function should be
         * callable from anywhere within the application without a risk of
         * deadlock.
         *
         * Returns:
         *  true if the thread is running, false if not.
         */
        final bit isRunning()
        {
            if( !m_addr )
                return false;
	        uint ecode = 0;
	        GetExitCodeThread( m_hndl, &ecode );
	        return ecode == STILL_ACTIVE;
        }

        /**
         * Waits for this thread to complete.
         *
         * Throws:
         *  ThreadError if the operation fails.
         */
        final void join()
        {
            if( WaitForSingleObject( m_hndl, INFINITE ) != WAIT_OBJECT_0 )
                throw new ThreadError( "Unable to join thread" );
        }

        /**
         * Suspends thread execution and loads stack and register data for use
         * by the suspending thread.  Registers should be stored in the order
         * they would typically appear when pushed onto the execution stack.
         * This function is reentrant, but resume must be called once for each
         * call to suspend.
         *
         * Throws:
         *  ThreadError if the thread is not currently running or if the thread
         *  can not be suspended.
         */
        final void suspend()
        {
            // NOTE: this operation uses the global shared lock for the following reasons:
            //         - typically, this function is called by suspendAll during a GC run,
            //           and it would be prohibitively expensive to have a separate synch
            //           block for every thread that needs to be suspended
            //         - implementing a correct lockless per-thread synchronization mechanism
            //           is too complicated to be worthwhile
            //         - this prevents the unlikely chance that a thread may try to suspend
            //           another thread that is beginning a GC run, thus deadlocking the app
            //       this implementation also relies on synch blocks being reentrant, as it
            //       is quite likely that a GC run will first lock the thread list and then
            //       attempt to pause each thread
            //
            // NOTE: it might be useful to use DBC to ensure that the suspending thread is
            //       the only thread that accesses this thread's stack pointers and such.
            synchronized( slock )
            {
                if( m_scount > 0 )
                {
                    ++m_scount;
                    return;
                }
                // NOTE: atempting to suspend a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && SuspendThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to suspend thread" );
                ++m_scount;

    	        CONTEXT context;
    	        context.ContextFlags = CONTEXT_INTEGER | CONTEXT_CONTROL;

    	        if( !GetThreadContext( m_hndl, &context ) )
    	            throw new ThreadError( "Unable to load thread context" );
                m_tstack = cast(void*) context.Esp;
                // edi,esi,ebp,esp,ebx,edx,ecx,eax
                m_reg[0] = context.Edi;
                m_reg[1] = context.Esi;
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                m_reg[4] = context.Ebx;
                m_reg[5] = context.Edx;
                m_reg[6] = context.Ecx;
                m_reg[7] = context.Eax;
                /+
                memcpy( &m_reg[0], &context.Edi, uint.sizeof * 2 );
                m_reg[2] = context.Ebp;
                m_reg[3] = context.Esp;
                memcpy( &m_reg[4], &context.Ebx, uint.sizeof * 4 );
                +/
            }
        }

        /**
         * Resumes thread execution and frees stack and register data.  This
         * function is reentrant and must be called once for every call to
         * suspend before the thread is resumed.
         *
         * Throws:
         *  ThreadError if the thread is not currently running or if the thread
         *  can not be resumed.
         */
        final void resume()
        {
            // NOTE: see comments in suspend for why this shared lock is used
            synchronized( slock )
            {
                if( m_scount < 1 )
                    return;
                if( m_scount > 1 )
                {
                    --m_scount;
                    return;
                }
                // NOTE: atempting to resume a non-running thread will throw an exception.
                //       is this reasonable?
                if( m_addr != GetCurrentThreadId() && ResumeThread( m_hndl ) == 0xFFFFFFFF )
                    throw new ThreadError( "Unable to resume thread" );
                --m_scount;

                m_tstack = m_bstack;
                memset( &m_reg[0], 0, uint.sizeof * m_reg.length );
            }
        }

        /**
         * Should be preceded by a call to suspend.
         *
         * Returns:
         *  A pointer to the logical top of the execution stack for this thread.
         *  Note that on some operating systems the address may be less than
         *  the address returned by stackBottom.
         */
        final void* stackTop()
        {
            return m_tstack;
        }

        /**
         * Should be preceded by a call to suspend.
         *
         * Returns:
         *  A pointer to the logical bottom of the execution stack for this
         *  thread.  Note that on some operating systems the address may be
         *  greater than the address returned by stackTop.
         */
        final void* stackBottom()
        {
            return m_bstack;
        }

        /**
         * Should be preceded by a call to suspend.
         *
         * Returns:
         *  A pointer to the beginning of an inclusive range of registers
         *  stored by suspend.  For example, if a system has registers X, Y, Z,
         *  and they are typically pushed onto the stack in that order, then
         *  this function will point to the value of X on systems where
         *  the stack grows up, or Z on systems where the stack grows down.
         */
        final void* regFirst()
        {
            return &m_reg[0];
        }

        /**
         * Should be preceded by a call to suspend.
         *
         * Returns:
         *  A pointer to the end of an inclusive range of registers stored by
         *  suspend.  For example, if a system has registers X, Y, Z, and they
         *  are typically pushed onto the stack in that order, then this
         *  function will point to the value of Z on systems where the stack
         *  grows up, or X on systems where the stack grows down.
         */
        final void* regLast()
        {
            return &m_reg[length-1];
        }

        /**
         * Adds a thread instance to the internal thread tracking mechanism
         * for tracking the main application thread.  Note that any
         * functionality related to garbage collection must operate before
         * this ctor or the module ctor has been called, and after the
         * respective dtors have been called.
         */
        static this()
        {
            sm_this = TlsAlloc();
            assert( sm_this != TLS_OUT_OF_INDEXES );

            Thread main   = new Thread();
            main.m_addr   = GetCurrentThreadId();
            main.m_hndl   = GetCurrentThreadHandle();
            main.m_bstack = getStackBottom();
            main.m_tstack = main.m_bstack;
            TlsSetValue( Thread.sm_this, main );

            add( main );
        }

        /**
         * Cleans up references to any threads that are no longer running.
         */
        static ~this()
        {
            foreach( Thread t; Thread )
            {
                if( !t.isRunning() )
                    remove( t );
            }
        }

        /**
         * Suspends the calling thread for at least the supplied time.
         *
         * Params:
         *  milliseconds = The minimum duration the calling thread should be
         *                 suspended.
         */
        static void sleep( uint milliseconds )
        {
            Sleep( milliseconds );
        }

        /**
         * Forces a context switch to occur away from the calling thread.
         */
        static void yield()
        {
            // NOTE: Sleep(1) is necessary because Sleep(0) does not give
            //       lower priority threads any timeslice, so looping on
            //       Sleep(0) could be resource-intensive in some cases.
            Sleep( 1 );
        }

        /**
         * Calls suspend on all running threads.
         */
        static void suspendAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.suspend();
                else
                    remove( curr );
            }
        }

        /**
         * Calls resume on all running threads.
         */
        static void resumeAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.resume();
                else
                    remove( curr );
            }
        }

        /**
         * The calling thread.
         *
         * Returns:
         *  The thread object representing the calling thread.  The result of
         *  deleting this object is undefined.
         */
        static Thread getThis()
        {
            // NOTE: this function may not be called until the module
            //       has been initialized.  this should not be an issue
            //       given the current implementation.
            return cast(Thread) TlsGetValue( sm_this );
        }

        /**
         * This function is not intended to be used by the garbage collector,
         * so memory allocation is allowed.
         *
         * Returns:
         *  An array containing references to all threads currently being
         *  tracked by the system.  The result of
         *  deleting any contained objects is undefined.
         */
        static Thread[] getAll()
        {
            synchronized( slock )
            {
                // NOTE: returning sm_all.values involves DMA
                //       which is unsafe if called from GC code
                return sm_all.values;
            }
        }

        /**
         * Operates on all threads currently tracked by the system.
         */
        static int opApply( int delegate( inout Thread ) dg )
        {
            synchronized( slock )
            {
                int ret = 0;

                foreach( Thread curr; sm_all )
                {
                    ret = dg( curr );
                    if( ret )
                        break;
                }
                return ret;
            }
        }

        /**
         * Indicates the number of local storage pointers available at program
         * startup.  It is recommended that this number be at least 64.
         */
        const uint LOCAL_MAX = 64;

        /**
         * Reserves a local storage pointer for use and initializes this
         * location to null for all running threads.
         *
         * Returns:
         *  A key representing the array offset of this memory location.
         */
        static uint createLocal()
        {
            synchronized( slock )
            {
                foreach( uint key, inout ubyte set; sm_local )
                {
                    if( !set )
                    {
                        foreach( Thread curr; sm_all )
                        {
                            curr.m_local[key] = null;
                        }
                        set = true;
                        return key;
                    }
                }
                throw new ThreadError( "No more local storage slots available" );
            }
        }

        /**
         * Marks the supplied key as available and sets the associated location
         * to null for all running threads.  It is assumed that any key passed
         * to this function is valid.  The result of calling this function for
         * a key which is still in use is undefined.
         *
         * Params:
         *  key = The key to delete.
         */
        static void deleteLocal( uint key )
        {
            synchronized( slock )
            {
                sm_local[key] = false;
                foreach( Thread curr; sm_all )
                {
                    curr.m_local[key] = null;
                }
            }
        }

        /**
         * Gets the data associated with the supplied key value.  It is assumed
         * that any key passed to this function is valid.
         *
         * Params:
         *  key = The location which holds the desired data.
         *
         * Returns:
         *  The data associated with the supplied key.
         */
        static void* getLocal( uint key )
        {
            return getThis().m_local[key];
        }

        /**
         * Stores the supplied value in the specified location.  It is assumed
         * that any key passed to this function is valid.
         *
         * Params:
         *  key = The location to store the supplied data.
         *  val = The data to store.
         *
         * Returns:
         *  A copy of the data which has just been stored.
         */
        static void* setLocal( uint key, void* val )
        {
            return getThis().m_local[key] = val;
        }

    protected:
        /**
         * This is the entry point for the newly invoked thread.  Default
         * behavior is to invoke the function or delegate passed on object
         * construction.  This function may be overridden to create custom
         * thread objects via subclassing.
         */
        void run()
        {
            switch( m_call )
            {
            case Call.FN:
                m_fn();
                break;
            case Call.DG:
                m_dg();
                break;
            default:
                break;
            }
        }

    private:
        static Object slock()
        {
            return Thread.classinfo;
        }

        static void add( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all[t.m_addr] = t;
            }
        }

        static void remove( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all.remove( t.m_addr );
            }
        }

        enum Call
        {
            NO,
            FN,
            DG
        }

        static ubyte[LOCAL_MAX] sm_local;
        static uint             sm_this;
        static Thread[uint]     sm_all; // indexed by addr

        void*[LOCAL_MAX]    m_local;
        HANDLE              m_hndl;
        uint                m_addr;
        void*               m_bstack,
                            m_tstack;
        uint[8]             m_reg; // edi,esi,ebp,esp,ebx,edx,ecx,eax
        int                 m_scount;
        Call                m_call;
        char[]              m_name;
        union
        {
            void function() m_fn;
            void delegate() m_dg;
        }
    }

    // Used for multiThreaded() below
    private static bit multiThreadedFlag = false;

    /**
     * This function is used to determine whether the the process is
     * multi-threaded.  Optimizations may only be performed on this
     * value if the programmer can guarantee that no path from the
     * enclosed code will start a thread.
     *
     * Returns:
     *  True if Thread.start() has been called in this process.
     */
    extern (C) bit multiThreaded()
    {
        return multiThreadedFlag;
    }

    /**
     * Identical to Thread.suspendAll().
     */
    extern (C) void suspendAllThreads()
    {
        Thread.suspendAll();
    }

    /**
     * Identical to Thread.resumeAll().
     */
    extern (C) void resumeAllThreads()
    {
        Thread.resumeAll();
    }

    private alias void delegate( void*, void* ) scanAllThreadsFn;

    /**
     * The main entry point for garbage collection.  The supplied delegate
     * will be passed ranges representing both stack and register values.
     *
     * Params:
     *  fn = The scanner function.  It should scan from p1 through p2 - 1.
     */
    extern (C) void scanAllThreads( scanAllThreadsFn fn )
    {
        foreach( Thread t; Thread )
        {
            // pointers should always refer to a valid range
            assert( t.stackTop && t.stackBottom );
            // stack grows down on x86 machines
            assert( t.stackTop <= t.stackBottom );
            fn( t.stackTop, t.stackBottom + 1 );

            // pointers should always refer to a valid range
            assert( t.regFirst && t.regLast );
            // regFirst should always be the min address
            assert( t.regFirst <= t.regLast );
	        fn( t.regFirst, t.regLast + 1 );
        }
    }
}
else
{
    private
    {
        import std.c.posix.semaphore;
        import std.c.posix.pthread;
        import std.c.posix.signal;
        import std.c.posix.unistd;
        import std.c.posix.time;
        import sys.linux.c.linux;

        //
        // entry point for POSIX threads
        //
        extern (C) void* threadFunc( void* arg )
        {
	        Thread  obj = cast(Thread) arg;
	        // maybe put an auto exception object here (using alloca) for OutOfMemoryError
	        // plus something to track whether an exception is in-flight?

            obj.m_bstack = getStackBottom();
	        obj.m_tstack = obj.m_bstack;
	        pthread_setspecific( sm_this, obj );
	        Thread.add( obj );
	        try
	        {
	            obj.run();
	        }
	        catch
	        {
	            // error should really print to stderr
	        }
	        Thread.remove( obj );
	        return null;
        }

        extern (C) void suspendHandler( int sig )
        in
        {
            assert( sig == SIGUSR1 );
        }
        body
        {
            Thread* obj = Thread.getThis();
            assert( obj );

            obj.m_tstack = getStackTop();
            asm
            {
                pushad;
            }
            assert( getStackTop() < obj.m_tstack );
            obj.m_freg = obj.m_tstack - 1;
            obj.m_lreg = getStackTop();

            // signal suspended
            // wait for wakeup signal

            obj.m_lreg = null;
            obj.m_freg = null;
	        asm
	        {
	            popad;
	        }
	        obj.m_tstack = obj.m_bstack;
        }
        /+
        extern (C) void resumeHandler( int sig )
        in
        {
            assert( sig == SIGUSR2 );
        }
        body
        {

        }
        +/
        //
        // NOTE: this may not work on all versions of linux,
        //       but apparently the Windows implementation
        //       does not work in Linux (according to DMD revs)
        //
        void* getStackBottom()
        {
            return __libc_stack_end;
        }

        void* getStackTop()
        {
        	asm
        	{   naked;
        	    mov EAX, ESP;
        	    ret;
        	}
        }
    }

    //
    // Thread class
    //
    class Thread
    {
        this()
        {
            m_call      = Call.NO;
        }

        this( void function() fn )
        {
            m_fn        = fn;
            m_call      = Call.FN;
        }

        this( void delegate() dg )
        {
            m_dg        = dg;
            m_call      = Call.DG;
        }

        ~this()
        in
        {
            assert( !isRunning() );
        }
        body
        {
            if( m_addr )
                remove( this );
            pthread_detach( m_addr );
        }

        final char[] name()
        {
            synchronized( this )
            {
                return m_name;
            }
        }

        final void name( char[] n )
        {
            synchronized( this )
            {
                m_name = n.dup;
            }
        }

        final void start()
        in
        {
            assert( !m_addr );
        }
        body
        {
            multiThreadedFlag = true;
            if( pthread_create( &m_addr, null, &threadFunc, this ) != 0 )
                throw new ThreadError( "Error creating thread" );
        }

        final bit isRunning()
        {
            // NOTE: possible implementations include registering a cleanup
            //       handler so we can detect thread cancel and manually
            //       tracking whether the thread function has exited or
            //       looking for a pthread function that can tell us this
            //       (possibly by checking for 'thread id invalid' return val)
        }

        final void join()
        {
            if( pthread_join( m_addr, null ) != 0 )
                throw exception( "Unable to join thread" );
        }

        final void suspend()
        {
            // NOTE: this operation uses the global shared lock for the following reasons:
            //         - typically, this function is called by suspendAll during a GC run,
            //           and it would be prohibitively expensive to have a separate synch
            //           block for every thread that needs to be suspended
            //         - implementing a correct lockless per-thread synchronization mechanism
            //           is too complicated to be worthwhile
            //         - this prevents the unlikely chance that a thread may try to suspend
            //           another thread that is beginning a GC run, thus deadlocking the app
            //       this implementation also relies on synch blocks being reentrant, as it
            //       is quite likely that a GC run will first lock the thread list and then
            //       attempt to pause each thread
            //
            // NOTE: it might be useful to use DBC to ensure that the suspending thread is
            //       the only thread that accesses this thread's stack pointers and such.
            synchronized( slock )
            {
                if( m_scount > 0 )
                {
                    ++m_scount;
                    return;
                }
                if( m_addr == pthread_self() )
                {
                    ++m_scount;

                    m_tstack = getStackTop();
    	            asm
    	            {
    	                pushad;
    	            }
    	            assert( getStackTop() < m_tstack );
    	            m_freg = m_tstack - 1;
    	            m_lreg = getStackTop();
                }
                else
                {
                    // NOTE: atempting to suspend a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR1 ) != 0 )
                        throw new ThreadError( "Unable to suspend thread" );
                    sem_wait( &m_suspend );
                    ++m_scount;
                }
            }
        }

        final void resume()
        {
            // NOTE: see comments in suspend for why this shared lock is used
            synchronized( slock )
            {
                if( m_scount < 1 )
                    return;
                if( m_scount > 1 )
                {
                    --m_scount;
                    return;
                }
                if( m_addr == pthread_self() )
                {
                    // pop registers off the stack before resuming
	                asm
	                {
	                    popad;
	                }
                }
                else
                {
                    // NOTE: atempting to resume a non-running thread will throw an exception.
                    //       is this reasonable?
                    if( pthread_kill( m_addr, SIGUSR2 ) != 0 )
                        throw new ThreadError( "Unable to resume thread" );
                }
                --m_scount;

                m_tstack = m_bstack;
                m_breg   = null;
                m_treg   = null;
            }
        }

        final void* stackTop()
        {
            return m_tstack;
        }

        final void* stackBottom()
        {
            return m_bstack;
        }

        final void* regFirst()
        {
            return m_freg;
        }

        final void* regLast()
        {
            return m_lreg;
        }

        static this()
        {
            int         status;
            sigaction_t sigusr1,
                        sigusr2;

            sigusr1.sa_flags   = 0;
            sigusr1.sa_handler = &suspendHandler;
            sigemptyset( &sigusr1.sa_mask );

            /+
            sigusr2.sa_flags   = 0;
            sigusr2.sa_handler = &resumeHandler;
            sigemptyset( &sigusr2.sa_mask );
            +/
            status = sigaction( SIGUSR1, &sigusr1, null );
            assert( status == 0 );
            /+
            status = sigaction( SIGUSR2, &sigusr2, null );
            assert( status == 0 );
            +/
            status = pthread_key_create( &sm_this, null );
            assert( status == 0 );

            Thread main   = new Thread();
            main.m_addr   = pthread_self();
            main.m_bstack = getStackBottom();
            main.m_tstack = main.m_bstack;

            status = pthread_setspecific( sm_this, main );
            assert( status == 0 );

            add( main );
        }

        static ~this()
        {
            foreach( Thread t; Thread )
            {
                if( !t.isRunning() )
                    remove( t );
            }
        }

        static void sleep( uint milliseconds )
        {
            if( uint.max / 1000 < milliseconds )
                sleep( milliseconds / 1000 );
            milliseconds %= 1000;
            if( milliseconds > 0 )
                usleep( milliseconds * 1000 );
        }

        static void yield()
        {
            sched_yield();
        }

        static void suspendAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.suspend();
                else
                    remove( curr );
            }
        }

        static void resumeAll()
        {
            foreach( Thread curr; Thread )
            {
                if( curr.isRunning() )
                    curr.resume();
                else
                    remove( curr );
            }
        }

        static Thread getThis()
        {
            // NOTE: this function may not be called until the module
            //       has been initialized.  this should not be an issue
            //       given the current implementation.  there is no
            //       precondition to test this because I have no idea
            //       whether zero is a legal value for an initialized
            //       pthread_key_t
            return cast(Thread) pthread_getspecific( sm_this );
        }

        static Thread[] getAll()
        {
            synchronized( slock )
            {
                // NOTE: returning sm_all.values involves DMA
                //       which is unsafe if called from GC code
                return sm_all.values;
            }
        }

        static int opApply( int delegate( inout Thread ) dg )
        {
            synchronized( slock )
            {
                int ret = 0;

                foreach( Thread curr; sm_all )
                {
                    ret = dg( curr );
                    if( ret )
                        break;
                }
                return ret;
            }
        }

        const uint LOCAL_MAX = 64;

        static uint createLocal()
        {
            synchronized( slock )
            {
                foreach( uint key, inout ubyte set; sm_local )
                {
                    if( !set )
                    {
                        foreach( Thread curr; sm_all )
                        {
                            curr.m_local[key] = null;
                        }
                        set = true;
                        return key;
                    }
                }
                throw new ThreadError( "No more local storage slots available" );
            }
        }

        static void deleteLocal( uint key )
        {
            synchronized( slock )
            {
                sm_local[key] = false;
                foreach( Thread curr; sm_all )
                {
                    curr.m_local[key] = null;
                }
            }
        }

        static void* getLocal( uint key )
        {
            return getThis().m_local[key];
        }

        static void* setLocal( uint key, void* val )
        {
            return getThis().m_local[key] = val;
        }

    protected:
        void run()
        {
            switch( m_call )
            {
            case Call.FN:
                m_fn();
                break;
            case Call.DG:
                m_dg();
                break;
            default:
                break;
            }
        }

    private:
        static Object slock()
        {
            return Thread.classinfo;
        }

        static void add( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all[t.m_addr] = t;
            }
        }

        static void remove( Thread t )
        in
        {
            assert( t.m_addr );
        }
        body
        {
            synchronized( slock )
            {
                sm_all.remove( t.m_addr );
            }
        }

        enum Call
        {
            NO,
            FN,
            DG
        }

        static ubyte[LOCAL_MAX]     sm_local;
        static pthread_key_t        sm_this;
        static Thread[pthread_t]    sm_all;

        pthread_t   m_addr;
        void*       m_bstack,
                    m_tstack,
                    m_freg,
                    m_lreg;
        int         m_scount;
        Call        m_call;
        char[]      m_name;
        union
        {
            void function() m_fn;
            void delegate() m_dg;
        }
    }

    // Used for multiThreaded() below
    private static bit multiThreadedFlag = false;

    extern (C) bit multiThreaded()
    {
        return multiThreadedFlag;
    }

    extern (C) void suspendAllThreads()
    {
        Thread.suspendAll();
    }

    extern (C) void resumeAllThreads()
    {
        Thread.resumeAll();
    }

    private alias void delegate( void*, void* ) scanAllThreadsFn;

    extern (C) void scanAllThreads( scanAllThreadsFn fn )
    {
        foreach( Thread t; Thread )
        {
            // pointers should always refer to a valid range
            assert( t.stackTop && t.stackBottom );
            // stack grows down on x86 machines
            assert( t.stackTop <= t.stackBottom );
            fn( t.stackTop, t.stackBottom + 1 );

            // pointers should always refer to a valid range
            assert( t.regFirst && t.regLast );
            // regFirst should always be the min address
            assert( t.regFirst <= t.regLast );
	        fn( t.regFirst, t.regLast + 1 );
        }
    }
}